function [DATA] = demo(examples)
%   Demo routine for usage of PAL.
%
%   Examples:
%
%   - demo('acoustic_wave_2d')   acoustic wave 2d problem, size 3540
%   - demo('r97')           a rotor dynamic problem, size 8652 
%   - demo('damped_beam')   damped beam problem, size 20000
%
%   Output DATA:
%       M, D, K, 
%	V, Lam, Res, info           % of PAL
%	V_DLIN, Lam_DLIN, Res_DLIN  % of DLIN	
%
%   See function pal for more information.
%
%   Revised:    April 22, 2016.

addpath ../;
doplot = true; % do eig plot

% Loading data
switch examples
    case 'acoustic_wave_2d'
    	load data/acoustic_2d_3540.mat; 
    	sigma = 170i;   % shift 
    	neig = 100;     % number of eigval
    	m = 3;          % Pade order 
    case 'r97'
    	load data/r97.mat;
    	sigma = 1e5i;   % shift 
    	neig = 100;     % number of eigval
    	m = 9;          % Pade order 
    case 'damped_beam'
        load data/damped_beam_20000.mat;
        sigma = 1e6i;   % shift
        neig = 100;     % number of eigval
        m = 9;          % Pade order 
    otherwise
        warning('Unexpected demo type.')
    return;
end
opts.disp = 0;

% Compute eig by direct linearization
tic;
[V1, Lam1, Res1] = dlin(M, C, K, sigma, neig); 
t1 = toc;

% Compute eig by pal
tic;
[V, Lam, Res, info] = pal(M, C, K, sigma, neig, m, opts); 
t = toc;

% Output data
if nargout > 0
    DATA.M = M; DATA.C = C; DATA.K = K;
    DATA.V = V; DATA.Lam = Lam; DATA.Res = Res; DATA.info = info;
    DATA.V_DLIN = V1; DATA.Lam_DLIN = Lam1; DATA.Res_DLIN = Res1; 
end

% Print summary
disp([' ---- Summary of the problem: ', examples, ' ----']);
disp(['    Problem size                     ', num2str(size(M,1))]);
disp(['    Daming matrix rank               ', num2str(info.rank)]);
disp(['    Pade order                       ', num2str(m)]);
disp(['    Required EigVals                 ', num2str(neig)]);
disp(['    Computing time of PAL            ', num2str(t),' s']);
disp(['    Computing time of DLIN           ', num2str(t1),' s']);
disp([' ------------------------------------------  ']);

% Plot eigenvalues and relative residual norms
figure;

subplot(1,2,1); % plot eigenvalues
plot(Lam1,'ok');
hold on; 
plot(Lam, '+r'); 
xlabel('real'); 
ylabel('imag'); 
legend('DLIN', ['PAL m =', num2str(m)]);
title('Eigenvalues');

subplot(1,2,2); % plot relative residuals
semilogy(sort(Res1,'ascend'), 'ok');
hold on; 
semilogy(sort(Res,'ascend'), '+r');
xlabel('Eig index'); ylabel('relative backward error');
legend('DLIN', ['PAL m =', num2str(m)]);
title('Relative Res.');

% END
return

