function [V, Lam, Res] = dlin(M, C, K, sigma, neig)
% [V, LAM, RES] = dlin(M, C, K, sigma, neig) computes neig eigenvalues
%   of the QEP 
%       Q(e) * v = (e^2 * M + e * C + K) * x = 0 
%   using direct linearization method, with PLV scaling.
%
%   On output, Lam is a vector of the neig computed eigenvalues, with each 
%   column of V being the corresponding (unitary) eigenvector. 
%   Res is a vector of the relative backward error of the eigenpairs: 
%   |Q(e) * v|_2 / (|e|^2 * |M|_1 + |e| * |C|_1 + |K|_1).
%   

n = length(M);

% Matrices norms
nm = norm(M,1);
nc = norm(C,1);
nk = norm(K,1);

% Computing scaling parameter (via PLV scaling)
omega = sqrt(nk / nm);
zeta = 2/(nk + omega * nc);

% Form scaled matrices.
Ms = omega^2 * M;
Cs = omega * ( C + 2 * sigma * M);
Qsigma = sigma^2 * M + sigma * C + K;

% Form MatVec operators of 
% [ 0       I   ]   [mu x] 
% [-Q\Ms    -Q\Cs]  [   x]

[LL UU PP] = lu(Qsigma, 0.1);
invqi = @(x) UU\(LL\(PP*x));
fun2 = @(x,y) -invqi( Ms * x + Cs * y );
fun = @(x) [x(n+1:end); fun2(x(1:n),x(n+1:end))];

% Compute the eigenvalues by eigs
opts.isreal = false;
[V1 RMU] = eigs(fun, 2 * n, neig, 'lm', opts);
mus = 1 ./ diag(RMU);
[mus, idx] = sort(mus, 'ascend');
Lam = omega * mus + sigma;
V1 = V1(:,idx);

% Select eigenvector according to magnitudes of mu 
n1 = length( find(abs(mus) < 1) );
V = [V1(n+1:end,1:n1), V1(1:n, n1+1:end) ];

% Compute relative backward error
nneig = length(Lam);
Res = zeros(nneig, 1);  
for j = 1:nneig
    e = Lam(j);
    v = V(:,j);
    v = v / norm(v);
    Res(j)  = norm( e^2 * M * v + e * C * v + K * v ) /...
        (abs(e)^2 * nm + abs(e) * nc + nk);
    V(:, j) = v;
end

return


